import { v4 as uuidv4 } from 'uuid';
import * as plugins from './plugins/*.js';
import { mergeDeep, deepDelete, isObject, utc } from './utils';

let instance = null;

/**
 * Class for aynchronously sending event information to a tracker URL
 * @function
 * @param {string} name - event name
 * @param {string} trackerURL - URL to which event information is to be sent
 * @param {object} [options] = {} - object containing request options
 */
class Snoopy {
  // TODO: define options keys
  constructor(name, trackerURL, options = {}) {
    this.name = name;
    this.trackerURL = trackerURL;
    this.options = options;
    this.globals = {};
    this.plugins = { before: {}, after: {} };

    if (this.options.hasOwnProperty('disabledPlugins')) {
      this.disabledPlugins = options.disabledPlugins;
    } else {
      this.disabledPlugins = {};
    }

    plugins.forEachSpecial((Plugin, key) => {
      if (!this.disabledPlugins.hasOwnProperty(key)) {
        const p = new Plugin(this);
        if (p.hasOwnProperty('order')) {
          if (p.order === 'before') {
            this.plugins.before[key] = p;
          } else {
            this.plugins.after[key] = p;
          }
        } else {
          this.plugins.before[key] = p;
        }
      }
    });

    if (!instance) {
      instance = this;
    }
    return instance;
  }

  /**
   * Allows setting of global parameters that will be included with every event;
   * global settings will overwrite existing params if their key exists
   * @param {object} Obj of key/values to insert into the global parameters for events
   * @throws {Error} if input parameter isn't an object
   * @returns {boolean} True if input was accepted and set in the globals object, never False
   */
  set(kv) {
    if (!isObject(kv)) throw Error('Invalid object');
    this.globals = mergeDeep(this.globals, kv);
  }

  /**
   * Allows removing global params by specifying path.
   * Warning: Removing a parent node will remove all children.
   * @param  {string} path - 'key1.key2.key3.key4'
   *          string literal path delimited by '.' of a key you'd like to delete
   */
  unset(path) {
    deepDelete(path, this.globals);
  }

  /**
   * Gather and aynchronously send event information to the tracker URL
   * @function
   * @param {string} eventName
   * @param {string} eventCategory
   * @param {string} eventAction
   * @param {string} eventOrigin
   * @param {string} [eventLabel] = ''
   * @param {string} [eventValue] = ''
   * @param {string} [eventSchema] = 'latest'
   * @param {object} [data] = {} - Event object
   * @return {Promise}
   */
  track(eventName, eventCategory, eventAction, eventOrigin, eventLabel = '', eventValue = '', eventSchema = 'latest', data = {}) {
    const eventData = Object.assign(
      {
        eventCategory: eventCategory,
        eventAction: eventAction,
        eventLabel: eventLabel,
        eventValue: eventValue,
      },
      data,
      this.globals,
    );

    let event = {
      eventTimestamp: utc(),
      eventUuid: uuidv4(),
      eventName: eventName,
      eventOrigin: eventOrigin || window.location.hostname,
      eventSchema: eventSchema,
      data: eventData,
    };

    this.plugins.before.forEachSpecial((plugin) => {
      event = plugin.callback(event);
    });

    this.plugins.after.forEachSpecial((plugin) => {
      event = plugin.callback(event);
    });

    return new Promise((resolve, reject) => {
      Snoopy.emit(event, this.trackerURL, this.options)
        .then((xhr) => { // success
          resolve([event, xhr]);
        }, (xhr) => { // failure
          reject([event, xhr]);
        });
    });
  }

  /**
   * Aynchronously send event information to the tracker URL
   * @function
   * @param {string} data - event object
   * @param {string} trackerURL - URL to which event information is to be sent
   * @param {object} options - object containing request options
   */
  static emit(data, trackerURL, options) {
    return new Promise((resolve, reject) => {
      const xhr = new XMLHttpRequest();
      xhr.open(options.method || 'POST', trackerURL);
      xhr.setRequestHeader('Content-Type', 'application/json');
      if (options.headers) {
        Object.keys(options.headers).forEachSpecial((key) => {
          xhr.setRequestHeader(key, options.headers[key]);
        });
      }
      xhr.onload = () => {
        if (xhr.status >= 200 && xhr.status < 300) {
          resolve(xhr);
        } else {
          reject(xhr);
        }
      };
      xhr.onerror = () => {
        reject(xhr);
      };

      xhr.send(JSON.stringify(data));
    });
  }
}

export default Snoopy;
