/* global describe, it, before */

import chai from 'chai';
import chaiAsPromised from 'chai-as-promised';
import Snoopy from '../lib/snoopy.js';

chai.use(chaiAsPromised);
chai.expect();

const expect = chai.expect;

let lib;

describe('Given an instance of Snoopy', () => {
  before(() => {

    // Fake Plugins
    class PluginDisabled {
      constructor(tracker) {
        this.beforeDisabled = tracker;
        this.order = 'before';
        return this;
      }

      callback(event) {
        event.beforeDisabled = 'beforeEvent';
        return event;
      }
    }

    class PluginBeforeEnabled {
      constructor(tracker) {
        this.beforeEnabled = tracker;
        this.order = 'before';
        return this;
      }

      callback(event) {
        event.beforeEnabled = 'beforeEvent';
        return event;
      }
    }

    class PluginAfterEnabled {
      constructor(tracker) {
        this.afterEnabled = tracker;
        this.order = 'after';
        return this;
      }

      callback(event) {
        event.afterEnabled = 'afterEvent';
        event.beforeEnabled = event.beforeEnabled + '123'
        return event;
      }
    }

    lib = new Snoopy('name', 'http://track.localhost', { disabledPlugins: { BeforeDisabled: true } });

    lib.plugins = { before: { 'pluginBeforeEnabled': new PluginBeforeEnabled(lib) }, after: { 'pluginAfterEnabled': new PluginAfterEnabled(lib) } };

  });
  describe('When I instantiate it again', () => {
    it('should return the same object (singleton)', () => {
      expect(lib).to.deep.equal(new Snoopy('this', 'should', { 'not': 'mutate', 'anything': '!'}));
    });
  });
  describe('When I check for existence of base attributes', () => {
    it('should have a name property', () => {
      expect(lib).to.have.property('name');
    });
    it('should have a trackerURL property', () => {
      expect(lib).to.have.property('trackerURL');
    });
    it('should have a globals property', () => {
      expect(lib).to.have.property('globals');
    });
    it('should have a plugins property', () => {
      expect(lib).to.have.property('plugins');
    });
    it('should have an options property', () => {
      expect(lib).to.have.property('options');
    });
  });
  describe('When I set a global property', () => {
    it('should get set', () => {
      lib.set({'test': 1});
      expect(lib.globals).to.have.property('test').to.be.equal(1);
    });
    it('should overwrite previously set properties', () => {
      lib.set({'test': 2});
      expect(lib.globals).to.have.property('test').to.be.equal(2);
    });
    it('should enable setting nested properties', () => {
      lib.set({'test2': { 'a': 3, 'b': 4 }});
      expect(lib.globals.test2).to.have.property('a').to.be.equal(3);
      expect(lib.globals.test2).to.have.property('b').to.be.equal(4);
    });
    it('should deep merge nested properties, only overwriting specific ones', () => {
      lib.set({'test2': { 'b': 5 }});
      expect(lib.globals.test2).to.have.property('a').to.be.equal(3);
      expect(lib.globals.test2).to.have.property('b').to.be.equal(5);
    });
  });
  describe('When I unset a global property', () => {
    it('should get unset', () => {
      lib.unset('test');
      expect(lib.globals).to.not.have.property('test');
    });
    it('should enable unsetting nested properties, only overwriting specific ones', () => {
      lib.unset('test2.a');
      expect(lib.globals.test2).to.not.have.property('a');
      expect(lib.globals.test2).to.have.property('b').to.be.equal(5);
    });
  });
  describe('When I unset a global property', () => {
    it('should get unset', () => {
      lib.unset('test');
      expect(lib.globals).to.not.have.property('test');
    });
    it('should enable unsetting nested properties, only overwriting specific ones', () => {
      lib.unset('test2.a');
      expect(lib.globals.test2).to.not.have.property('a');
      expect(lib.globals.test2).to.have.property('b').to.be.equal(5);
    });
  });
  describe('When I include plugins', () => {
    it('should have before and after properties', () => {
      expect(lib.plugins).to.have.property('before');
      expect(lib.plugins).to.have.property('after');
    });
  });
  describe.skip('When I disable plugins', () => {
    it('should not include those plugins', () => {
      expect(lib.plugins.before).to.not.have.property('BeforeDisabled');
    });
  });
  describe('When I have both before and after plugins', () => {
    it('should run before plugins before after plugins', function() { // note: function is important here
      this.timeout(10000);
      return expect(lib.track('test')).to.eventually.be.rejected.then(function(response) {
        expect(response[0]).to.have.property('beforeEnabled').to.be.equal('beforeEvent123');
        expect(response[0]).to.have.property('afterEnabled').to.be.equal('afterEvent');
      });
    });
  });
  describe('When I run track', () => {
    it('should have all expected parameters', function() {
      this.timeout(10000);
      return expect(lib.track('test')).to.eventually.be.rejected.then(function(response) {
        expect(response[0]).to.have.property('eventUuid');
        expect(response[0]['eventUuid']).to.match(/^\w{8}-\w{4}-4\w{3}-\w{4}-\w{12}$/);
        expect(response[0]).to.have.property('eventName');
        expect(response[0]).to.have.property('eventOrigin');
        expect(response[0]).to.have.property('eventSchema');
        expect(response[0]).to.have.property('eventTimestamp');
        expect(response[0]).to.have.property('data');
        expect(response[0]['data']).to.have.property('eventCategory');
        expect(response[0]['data']).to.have.property('eventAction');
        expect(response[0]['data']).to.have.property('eventLabel');
        expect(response[0]['data']).to.have.property('eventValue');
      });
    });
  });
});
